import { getDevOpsCore } from '@thatdevops/core/devopsCore'
import getPkg from '@thatdevops/core/pkgman/getPkg'
import getProjectPkgs from '@thatdevops/core/pkgman/getProjectPkgs'
import writePkg from '@thatdevops/core/pkgman/writePkg'
import echo from '@thatdevops/core/utils/echo'
import chunk from 'lodash/chunk'
import { DateTime } from 'luxon'
import path from 'path'
import BumpType from './BumpType'
import doCalcNewVersions from './doCalcNewVersions'
import doCommit from './doCommit'
import doCommitPush from './doCommitPush'
import doFindChangedPackages from './doFindChangedPackages'
import doFindEffectedPackages from './doFindEffectedPackages'
import doUpdateDependencies from './doUpdateDependencies'
import doYarnPublish from './doYarnPublish'

export async function doCommitPushPublish(props?: {
	force?: boolean
	bumpType?: BumpType
	commitMessage?: string
}) {
	// Get properties
	const { rootPath, parallel = 0 } = getDevOpsCore()
	let { force = false, commitMessage, bumpType = BumpType.patch } = props ?? {}

	if (!commitMessage) commitMessage = `Publish ${DateTime.now().toISODate()}`

	// Is the project publishable?
	const { publishable } = getDevOpsCore()
	if (!publishable) {
		echo('Publishing not allowed (set the dev ops "publishable" flag')
		return
	}

	// Pre commit
	await doCommit(`Pre ${commitMessage}`)

	// Find packages that have changes since their last publish
	echo('Finding changed packages')
	let changedPkgs = force ? await getProjectPkgs() : await doFindChangedPackages()

	// If no changes, then nothing to do
	if (!changedPkgs.length) {
		echo('No changes to publish')
		return
	}

	// Find effected packages
	const effectedPkgs = await doFindEffectedPackages(changedPkgs)

	// Calculate new versions for changed packages
	const newVersions = doCalcNewVersions(effectedPkgs, bumpType)

	// Update any dependencies that have changed
	echo('Updating related dependencies')
	await doUpdateDependencies(effectedPkgs, newVersions)

	// Publish effected packages
	echo('Publishing effected packages')
	if (!!parallel) {
		const chunks = chunk(effectedPkgs, parallel)
		for (const chunki of chunks) {
			await Promise.all(chunki.map(c => doYarnPublish(c, newVersions[c.name])))
		}
	} else {
		for (const pkg of effectedPkgs) {
			await doYarnPublish(pkg, newVersions[pkg.name])
		}
	}

	// Update devops package
	echo('Updating DevOps package')
	const devopsPkg = getPkg(path.join(rootPath, 'devops'))
	await doUpdateDependencies([devopsPkg], newVersions)

	// Commit and push
	await doCommitPush(commitMessage)

	// Set publish dates
	echo('Setting publish dates')
	const rootPkg = getPkg(rootPath)
	rootPkg.json.devops = rootPkg.json.devops || {}
	rootPkg.json.devops.publishdates = rootPkg.json.devops.publishdates || {}
	const now = DateTime.now().plus({ minute: 1 }).toISO()
	for (const pkg of effectedPkgs) {
		rootPkg.json.devops.publishdates[pkg.json.name] = now
	}
	writePkg(rootPkg)
}

export default doCommitPushPublish
